% =========================================================================
% Title       : Error Simulation Environment for large-scale MIMO downlink
% File        : err_sim.m
% -------------------------------------------------------------------------
% Description :
%   Main simulation file for large-scale MIMO downlink
% ------------------------------------------------------------------------- 
% Revisions   :
%   Date       Version  Author  Description
%   07-apr-13  1.3      studer  cleanup
%   01-dec-08  1.2      studer  major cleanup and speedup
%   14-sep-06  1.1      studer  enhanced with iterative MIMO decoding
%   10-aug-06  1.0      studer  adapted from ofdm_sim environment
% -------------------------------------------------------------------------
%   (C) 2013 Christoph Studer (e-mail: studer@rice.edu)     
% =========================================================================

function err_sim(TxRx,RunID)
  
  % -- initialization 
  [TxRx,Result,Stat] = init(TxRx,RunID);
  K_list = 1:length(TxRx.Sim.SNR_list);
    
  % -- main simulation loop (packets)
  tic;
  for pack=1:TxRx.Sim.nr_of_packets

    % -- generate channel matrices according to specified channel model
    HN = channel(TxRx); 

    % -- transmit OFDM symbols
    PERRORunc = zeros(length(TxRx.Sim.SNR_list),1);
    PERROR = zeros(length(TxRx.Sim.SNR_list),1);    
    for symb=1:TxRx.Sim.nr_of_symbols
    
      % -- transmitter: generates bits, maps to symbol vectors, 
      % -- and transmits symbol vectors over (noiseless) channel 
      [data_bin,data_bin_coded,rxN,Stat,nscale] = tx(TxRx,HN,Stat);
            
      % -- generate noise realization
      noise = sqrt(0.5)*(randn(TxRx.Nrx,TxRx.Nused)+1i*randn(TxRx.Nrx,TxRx.Nused)); %%%%%%%%%%%%%%%%%%% CHECK!!!
    
      
      for k=K_list
        % -- compute SNR with instantaneous re-scaling
        SNR = TxRx.Sim.SNR_list(k);
        N0 = (nscale/SNR);
           
        % -- add Gaussian noise
        yN = rxN + noise*sqrt(N0); 

         
        % -- wrapper for detectors
        [LLR_E1,~] = detector(TxRx,yN,HN,N0);  
     
        % -- wrapper for channel decoders
        [~,data_bin_hat] = decoder(TxRx,LLR_E1);           
     
        errs_uncoded = sum(abs((LLR_E1>0)-data_bin_coded)');
        errs_coded = sum(abs(data_bin_hat-data_bin)');
        
          % -- compute coded BER/FER
          %tmp = sum(sum(abs(data_bin_hat-data_bin)))
          %[ complexity , tmp ] 
          Result.BERunc(k,:) = Result.BERunc(k,:) + errs_uncoded;
          Result.SERunc(k,:) = Result.SERunc(k,:) + (errs_uncoded>0);   
          if (sum(errs_uncoded)>0)
            PERRORunc(k) = 1;
          end
          
          Result.BER(k,:) = Result.BER(k,:) + errs_coded;
          Result.SER(k,:) = Result.SER(k,:) + (errs_coded>0); 
          if (sum(errs_coded)>0)
            PERROR(k) = 1;
          end                     
              
      end % end SNR loop

    end % end symbol loop
 
    % -- count packet errors
    Result.PERunc = Result.PERunc + PERRORunc;
    Result.PER = Result.PER + PERROR;
    
    Result.Npackets(K_list) = Result.Npackets(K_list) + 1;
    K_list(Result.PERunc(K_list)>TxRx.Sim.MaxPER) = []; %%% stopping criterion for uncoded transmission...
    
    % -- save intermediate results
    Result.total_time = Result.total_time + toc;
    save_results(TxRx,RunID,Result,pack,Stat);
    
    % -- show # of elapsed packets
    disp(sprintf('\nPacket %i of %i transmitted',pack,TxRx.Sim.nr_of_packets))
      
  end % end packet loop

  % -- simulation completed!
  disp(' =====================================================');
  disp(sprintf(' ==================== Total Simulation Time: %f sec. ',Result.total_time));
  disp(' =====================================================');

  % -- store FINAL results to disk
  save_results(TxRx,RunID,Result,TxRx.Sim.nr_of_packets,Stat);
  
return

% -------------------------------------------------------------------------
% -- save (intermediate/final) results to disk
% -------------------------------------------------------------------------
function save_results(TxRx,RunID,Res,NrOfSimPackets,Stat) 
  % -- compose simulation results structure
  Results.TxRx = TxRx;
  Results.Npackets = Res.Npackets;  
  Results.RunName = TxRx.Sim.name;
  Results.RunID = RunID;
  Results.simulationTime = Res.total_time; 
  Results.NrOfSimPackets = NrOfSimPackets; % store actual number of simulated channels
  if (NrOfSimPackets==TxRx.Sim.nr_of_packets)
    Results.Final = 'yes';
  else
    Results.Final = 'no';
  end
  % -- for each iteration
  Results.BERunc = Res.BERunc./(Res.Npackets*ones(1,TxRx.Nrx)*(TxRx.Sim.nr_of_symbols*TxRx.Nused*TxRx.Modulation_order));  
  Results.SERunc = Res.SERunc./(Res.Npackets*ones(1,TxRx.Nrx)*TxRx.Sim.nr_of_symbols);
  Results.PERunc = Res.PERunc./Res.Npackets;
  
  Results.BER = Res.BER./(Res.Npackets*ones(1,TxRx.Nrx)*(TxRx.Sim.nr_of_symbols*TxRx.Nused*TxRx.Modulation_order*TxRx.Code.Rate));  
  Results.SER = Res.SER./(Res.Npackets*ones(1,TxRx.Nrx)*TxRx.Sim.nr_of_symbols);
  Results.PER = Res.PER./Res.Npackets;
  
  Results.Stat = Stat;
  
  % -- store Results structure to a separate file
  Results.FileName = sprintf('results/%s_%d.mat',TxRx.Sim.name,RunID);
  save(Results.FileName,'Results');
  
return