% =========================================================================
% Title       : Fast iterative truncation algorithm (FITRA)
% File        : FITRA.m
% -------------------------------------------------------------------------
% Description :
%
%   First-order algorithm for approximating l_infty-norm regularized 
%   least-squares problems. The algorithm builds on FISTA as described 
%   by A. Beck and M. Teboulle 2009 and contains a fast proximal-map
%   computation step to reduce the computational complexity.
%
% Solves:
%
%   xhat = argmin_x { lambda*||x||_inf + ||A*x-b||^2_2 }
%
% Usage:
%
%   xhat = FITRA(A,At,b,N,lambda,max_iter)
%
%   A        : function handle for A*x operation
%   At       : function handle for A^t*y operation
%   b        : signal to be represented
%   N        : dimension of the signal to be represented
%   lambda   : regularization parameter
%   max_iter : maximum number of iterations
%   xhat     : democratic representation (with low PAR)
%
% ------------------------------------------------------------------------- 
% Revisions   :
%   Date       Version  Author  Description
%   07-apr-12  1.2      studer  cleanup
%   01-may-12  1.1      studer  faster proximal map
%   14-dec-11  1.0      studer  file created
% -------------------------------------------------------------------------
% (C) 2013 Christoph Studer (e-mail: studer@rice.edu)
% =========================================================================

function [xhat,i] = FITRA(A,At,b,N,lambda,max_iter,tol)

  % -- initialization
  xkp = zeros(N,1);
  yk = xkp;
  tk = 1;  
   
  % -- calculate Lipschitz constant (backtracking might be faster)
  L = 2*maxeig(A,At,N,1e-9); %L = 2*max(svd(A))^2; %L = 2*max(svd(A'*A)); 
  
  % -- begin main (outer) iteration
  normdiff = inf;
  normx = 0;
  i=0;
  while (i<max_iter)&&(normdiff>normx*tol)
    i = i + 1;
    
    % -- gradient step
    w = yk - 2/L*At(A(yk)-b);
    
    % -- N*log(N) method to compute truncation level
    wabs = abs(w);
    ws = (cumsum(sort(wabs,'descend'))- lambda/L)./(1:N)';
    alphaopt = max(ws);
    if alphaopt>0 
      xk = min(abs(w),alphaopt).*sign(w); % truncation step
    else
      xk = w; % do not truncate
    end           
      
    % -- perform FISTA-like iterations    
    tkn = (1+sqrt(1+4*tk^2))/2;
    ykn = xk + (tk-1)/tkn*(xk-xkp);        
    normdiff = norm(yk-ykn,2);
    normx = norm(yk,2);     
    xkp = xk;
    yk = ykn;
    tk = tkn;
    
  end
  
  xhat = xk;

end

% -- find max. eigenvalue of A'*A using power iterations (slow)
function [val,iters] = maxeig(A,At,N,tol)
  iters = 0;
  bk = randn(N,1); 
  bk = bk/norm(bk,2);
  err = inf;
  while err>tol
    iters = iters + 1;
    bkold = bk;
    tmp = At(A(bk));
    val = norm(tmp);    
    bk = tmp/val;    
    err = norm(bkold-bk,2)^2/N;
  end
end
