% =========================================================================
% Title       : Main file for infty-norm based PAPR reduction
% File        : PAPRed.m
% -------------------------------------------------------------------------
% Description :
%   Performs l_infty-norm based PAPR reduction. Calls the FITRA algorithm
%   to compute the signal representation with small PAPR. 
% ------------------------------------------------------------------------- 
% Revisions   :
%   Date       Version  Author  Description
%   07-apr-13  1.0      studer  file created
% -------------------------------------------------------------------------
%   (C) 2013 Christoph Studer (e-mail: studer@rice.edu)     
% =========================================================================

function [xN,tN,OPR] = PAPRed(TxRx,HN,sN)

  % -- variable initialization

  TxRx.ToneMapC = setdiff(1:TxRx.N,TxRx.ToneMap); % unused tones
  TxRx.NusedC = TxRx.N-TxRx.Nused; % number of unused tones

  
  % -- define operators
  Afun_op = @(tvec) Afun(TxRx,HN,tvec);  
  Atfun_op = @(svec) Atfun(TxRx,HN,svec); 

  % -- convert TX symbols to real-valued vector
  scpx = sN(:);
  svec = [ real(scpx) ; imag(scpx) ; zeros(TxRx.Ntx*TxRx.NusedC*2,1) ];
  
  
  % -- run FITRA algorithm 
  M = TxRx.Nused*TxRx.Nrx*2 + TxRx.Ntx*TxRx.NusedC*2;
  N = TxRx.N*TxRx.Ntx*2;
  
  tic
  that = FITRA(Afun_op,Atfun_op,svec,N,TxRx.ShapingAlg.lambda,TxRx.ShapingAlg.iter,TxRx.ShapingAlg.tol);
  toc
  
  % -- convert result to frequecy domain
  
  tcpx = that(1:TxRx.N*TxRx.Ntx,1) + 1i*that(TxRx.N*TxRx.Ntx+1:end,1);
  tN = zeros(TxRx.N,TxRx.Ntx);
  tN(:) = tcpx;
  Atmp = 1/sqrt(TxRx.N)*fft(tN); % orthonormal FT
  xN = Atmp(TxRx.ToneMap,:).'; % subsample and transpose
  
  % -- out-of-band power OPR
  OPR = length(TxRx.ToneMap)/length(TxRx.ToneMapC)*sum(sum(abs(Atmp(TxRx.ToneMapC,:).').^2))/sum(sum(abs(Atmp(TxRx.ToneMap,:).').^2));

end

% -- functions that implement the action of OFDM processing and the
% -- channels in frequency domain

function svec = Afun(TxRx,HN,tvec)
  tcpx = tvec(1:TxRx.N*TxRx.Ntx,1) + 1i*tvec(TxRx.N*TxRx.Ntx+1:end,1);
  A = zeros(TxRx.N,TxRx.Ntx);
  A(:) = tcpx;
  Atmp = 1/sqrt(TxRx.N)*fft(A); 
  Xn = Atmp(TxRx.ToneMap,:).'; % subsample and transpose
  Qn = Atmp(TxRx.ToneMapC,:);
  qconst = Qn(:);
  scpx = zeros(TxRx.Nrx*TxRx.Nused,1);
  for n=1:TxRx.Nused
    scpx((n-1)*TxRx.Nrx+1:n*TxRx.Nrx,1) = HN(:,:,TxRx.ToneMap(n))*Xn(:,n);    
  end
  svec = [ real(scpx) ; imag(scpx) ; real(qconst) ; imag(qconst) ];
end

function tvec = Atfun(TxRx,HN,svec)
  scpx = svec(1:TxRx.Nused*TxRx.Nrx,1) + 1i*svec(TxRx.Nused*TxRx.Nrx+1:TxRx.Nused*TxRx.Nrx*2,1);  
  qconst = svec(TxRx.Nused*TxRx.Nrx*2+1:TxRx.Nused*TxRx.Nrx*2+TxRx.Ntx*TxRx.NusedC,1) ...
                 + 1i*svec(TxRx.Nused*TxRx.Nrx*2+TxRx.Ntx*TxRx.NusedC+1:end,1);
  Qn = zeros(TxRx.NusedC,TxRx.Ntx);
  Qn(:) = qconst;
  Xn = zeros(TxRx.Ntx,TxRx.Nused);
  for n=1:TxRx.Nused
    Xn(:,n) = HN(:,:,TxRx.ToneMap(n))'*scpx((n-1)*TxRx.Nrx+1:n*TxRx.Nrx,1);
  end
  A = zeros(TxRx.N,TxRx.Ntx);
  A(TxRx.ToneMap,:) = Xn.'; % transpose!
  A(TxRx.ToneMapC,:) = Qn;
  Atmp = sqrt(TxRx.N)*ifft(A);
  tcpx = Atmp(:);
  tvec = [ real(tcpx) ; imag(tcpx) ];
end

