% =========================================================================
% Title       : Large-scale MIMO transmitter with precoding
% File        : tx.m
% -------------------------------------------------------------------------
% Description :
%   This file generates random bits, applies forward error-correction with
%   a convolutional encoder (note: convenc reqiures communication toolbox).
%   Furthermore, interleaving and modulation is performed and the symbol
%   vectors are transmitted over the channel matrices (without noise).
% ------------------------------------------------------------------------- 
% Revisions   :
%   Date       Version  Author  Description
%   07-apr-13  1.0      studer  file created
% -------------------------------------------------------------------------
%   (C) 2013 Christoph Studer (e-mail: studer@rice.edu)     
% =========================================================================

function [data_bin,data_bin_coded_stream,rxN,Stat,nscale] = tx(TxRx,HN,Stat)

  % -- generate random bits and apply forward error correction   
  data_bin = round(rand(TxRx.Nrx,ceil(TxRx.Nused*TxRx.Modulation_order*TxRx.Code.Rate)));
  data_bin(1:TxRx.Nrx,[size(data_bin,2)+1-(TxRx.Code.K-1)-TxRx.Code.PADbits:size(data_bin,2)]) = 0;
  
  % -- separate enconding for each terminal
  for nrx=1:TxRx.Nrx
    data_bin_coded(nrx,:) = convenc(data_bin(nrx,:),TxRx.Code.trellis); % requires communication toolbox
  
    % -- puncturing
    data_bin_punctured(nrx,:) = data_bin_coded(nrx,TxRx.Code.Puncturing.Index);
  
    % -- interleaving
    data_bin_coded_stream(nrx,:) = data_bin_punctured(nrx,TxRx.Code.InterleaverPerm(nrx,:));

  end
  
  % -- modulation (prepare symbol vectors)
  for nrx=1:TxRx.Nrx
    tmp = reshape(data_bin_coded_stream(nrx,:),TxRx.Modulation_order,TxRx.Nused);
    tmp_dec = bi2de(tmp.').';
    sN(nrx,:) = TxRx.Constellations_norm(tmp_dec+1);
  end

  % -- shape the transmit signals
  % -- using a precoding/combining/signal-shaping method
  % -- *** assumes perfect knowledge of H ***
  xN = zeros(TxRx.Ntx,TxRx.Nused);      
  switch (TxRx.Shaping)
    case 'LS', % -- ZF or least-squares (LS) precoding
      tic
      for n=1:TxRx.Nused   
        xN(:,n) = pinv(HN(:,:,TxRx.ToneMap(n)))*sN(:,n);
      end      
      Q = zeros(TxRx.N,TxRx.Ntx);
      Q(TxRx.ToneMap,:) = xN.';      
      tN = sqrt(TxRx.N)*ifft(Q);
      toc
      OPR = 0;
      
    case 'MF', % -- matched-filter (MF) combining
      tic
      for n=1:TxRx.Nused  
        G = HN(:,:,TxRx.ToneMap(n))'*diag(1./diag(HN(:,:,TxRx.ToneMap(n))*HN(:,:,TxRx.ToneMap(n))'));
        xN(:,n) = G*sN(:,n);
      end      
      Q = zeros(TxRx.N,TxRx.Ntx);
      Q(TxRx.ToneMap,:) = xN.';      
      tN = sqrt(TxRx.N)*ifft(Q);
      thatcpx = tN(:);
      that = [ real(thatcpx) ; imag(thatcpx) ];
      toc      
      OPR = 0;
                      
    case 'PMP', % -- Precoding, modulation and PAR reduction (PMP)
      [xN,tN,OPR] = PAPRed(TxRx,HN,sN);
      
    otherwise,
      error('TxRx.Shaping method not defined')
  end
   
  % == extract some statistics
  
  PAR = TxRx.N*2*max(abs([real(tN);imag(tN)])).^2./sum(abs(tN).^2); 
  Stat.PAR = [Stat.PAR PAR];  
  Stat.OPR = [Stat.OPR OPR];

  % -- used for normalizing the average transmit power per symbol
  nscale = (norm(xN,'fro')^2/TxRx.Nused);
  
  % -- transmit over the noiseless channel
  rxN = zeros(TxRx.Nrx,TxRx.Nused);
  for n=1:TxRx.Nused
    rxN(:,n) = HN(:,:,TxRx.ToneMap(n))*xN(:,n);
  end 
      
end
  